"""Component implementations in various backends.

Modules in this package and associated subpackages must never be imported into the core
Bitfount package unless a check is performed to ensure the given backend exists.
Every subpackage in this package must have an `__init__.py` file that defines `__all__`.
"""
