# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: bitfount/federated/transport/protos/messages.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="bitfount/federated/transport/protos/messages.proto",
    package="messages",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n2bitfount/federated/transport/protos/messages.proto\x12\x08messages"\x9c\x01\n\x07PodData\x12\x0f\n\x07podName\x18\x01 \x01(\t\x12\x11\n\tprocessor\x18\x02 \x01(\t\x12\r\n\x05podOS\x18\x03 \x01(\t\x12\x10\n\x08\x63puCount\x18\x04 \x01(\r\x12\x18\n\x10totalMemoryBytes\x18\x05 \x01(\x04\x12\x10\n\x08gpuCount\x18\x06 \x01(\r\x12\x14\n\x07gpuName\x18\x07 \x01(\tH\x00\x88\x01\x01\x42\n\n\x08_gpuName"u\n\x0c\x42itfountTask\x12\x15\n\rpodIdentifier\x18\x01 \x01(\t\x12\x1a\n\rencryptedTask\x18\x03 \x01(\x0cH\x00\x88\x01\x01\x12\x14\n\x07taskURL\x18\x04 \x01(\x0cH\x01\x88\x01\x01\x42\x10\n\x0e_encryptedTaskB\n\n\x08_taskURL"6\n\rBitfountTasks\x12%\n\x05tasks\x18\x01 \x03(\x0b\x32\x16.messages.BitfountTask"\xb9\x01\n\x14\x43ommunicationDetails\x12\x11\n\tmailboxId\x18\x01 \x01(\t\x12H\n\rpodMailboxIds\x18\x02 \x03(\x0b\x32\x31.messages.CommunicationDetails.PodMailboxIdsEntry\x12\x0e\n\x06taskId\x18\x03 \x01(\t\x1a\x34\n\x12PodMailboxIdsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\xbb\x07\n\x0f\x42itfountMessage\x12:\n\x0bmessageType\x18\x01 \x01(\x0e\x32%.messages.BitfountMessage.MessageType\x12\x0c\n\x04\x62ody\x18\x02 \x01(\x0c\x12\x11\n\trecipient\x18\x03 \x01(\t\x12\x1a\n\x12recipientMailboxId\x18\x04 \x01(\t\x12\x13\n\x06sender\x18\x05 \x01(\tH\x00\x88\x01\x01\x12\x17\n\x0fsenderMailboxId\x18\x06 \x01(\t\x12\x11\n\ttimestamp\x18\x07 \x01(\t\x12\x43\n\rpodMailboxIds\x18\x08 \x03(\x0b\x32,.messages.BitfountMessage.PodMailboxIdsEntry\x12\x1a\n\rreceiptHandle\x18\t \x01(\tH\x01\x88\x01\x01\x12\x13\n\x06taskId\x18\n \x01(\tH\x02\x88\x01\x01\x1a\x34\n\x12PodMailboxIdsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\x99\x04\n\x0bMessageType\x12\r\n\tUNDEFINED\x10\x00\x12\x16\n\x12\x41LGORITHM_EXCHANGE\x10\x01\x12\x0f\n\x0bJOB_REQUEST\x10\x02\x12\x0e\n\nJOB_ACCEPT\x10\x03\x12\x0e\n\nJOB_REJECT\x10\x04\x12\x13\n\x0fTRAINING_UPDATE\x10\x05\x12\x15\n\x11TRAINING_COMPLETE\x10\x06\x12\x16\n\x12\x45VALUATION_RESULTS\x10\x07\x12\x10\n\x0cSECURE_SHARE\x10\x08\x12\x14\n\x10MODEL_PARAMETERS\x10\t\x12\x0e\n\nTASK_ABORT\x10\n\x12 \n\x1cLARGE_OBJECT_STORAGE_REQUEST\x10\x0b\x12\x10\n\x0cKEY_EXCHANGE\x10\x0c\x12\x0c\n\x08POD_DATA\x10\r\x12\x10\n\x0cSAML_REQUEST\x10\x0e\x12\x11\n\rSAML_RESPONSE\x10\x0f\x12\x0f\n\x0bLOG_MESSAGE\x10\x10\x12\x14\n\x10TRAINING_METRICS\x10\x11\x12\x11\n\rTASK_COMPLETE\x10\x12\x12\x12\n\x0eOIDC_CHALLENGE\x10\x13\x12\x1a\n\x16OIDC_AFC_PKCE_RESPONSE\x10\x14\x12\x1d\n\x19OIDC_DEVICE_CODE_RESPONSE\x10\x15\x12\x10\n\x0cONLINE_CHECK\x10\x16\x12\x13\n\x0fONLINE_RESPONSE\x10\x17\x12\x0f\n\x0bPSI_DATASET\x10\x18\x12\x0e\n\nTASK_START\x10\x19\x42\t\n\x07_senderB\x10\n\x0e_receiptHandleB\t\n\x07_taskId"i\n\x0f\x41\x63knowledgement\x12\x11\n\tmailboxId\x18\x01 \x01(\t\x12\x15\n\rreceiptHandle\x18\x03 \x01(\t\x12\x1a\n\rdeleteMailbox\x18\x04 \x01(\x08H\x00\x88\x01\x01\x42\x10\n\x0e_deleteMailbox"\x11\n\x0fSuccessResponse"\x85\x01\n\x13LargeStorageRequest\x12\x17\n\x0fsenderMailboxId\x18\x01 \x01(\t\x12\x14\n\x07podName\x18\x02 \x01(\tH\x00\x88\x01\x01\x12\x13\n\x0b\x63ontentSize\x18\x03 \x01(\r\x12\x13\n\x06taskId\x18\x04 \x01(\tH\x01\x88\x01\x01\x42\n\n\x08_podNameB\t\n\x07_taskId"_\n\x17LargeStorageTaskRequest\x12\x0f\n\x07podName\x18\x01 \x01(\t\x12\x13\n\x0b\x63ontentSize\x18\x02 \x01(\r\x12\x13\n\x06taskId\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\t\n\x07_taskId"\xb1\x01\n\x0f\x42lobStorageData\x12\x11\n\tuploadUrl\x18\x01 \x01(\t\x12\x13\n\x0b\x64ownloadUrl\x18\x02 \x01(\t\x12\x41\n\x0cuploadFields\x18\x03 \x03(\x0b\x32+.messages.BlobStorageData.UploadFieldsEntry\x1a\x33\n\x11UploadFieldsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x32\xe4\x03\n\x0eMessageService\x12<\n\nPodConnect\x12\x11.messages.PodData\x1a\x19.messages.SuccessResponse"\x00\x12O\n\x12SetupTaskMailboxes\x12\x17.messages.BitfountTasks\x1a\x1e.messages.CommunicationDetails"\x00\x12L\n\x12\x41\x63knowledgeMessage\x12\x19.messages.Acknowledgement\x1a\x19.messages.SuccessResponse"\x00\x12Q\n\x12GetBitfountMessage\x12\x1e.messages.CommunicationDetails\x1a\x19.messages.BitfountMessage"\x00\x12M\n\x13SendBitfountMessage\x12\x19.messages.BitfountMessage\x1a\x19.messages.SuccessResponse"\x00\x12S\n\x15GetLargeObjectStorage\x12\x1d.messages.LargeStorageRequest\x1a\x19.messages.BlobStorageData"\x00\x62\x06proto3',
)


_BITFOUNTMESSAGE_MESSAGETYPE = _descriptor.EnumDescriptor(
    name="MessageType",
    full_name="messages.BitfountMessage.MessageType",
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name="UNDEFINED",
            index=0,
            number=0,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="ALGORITHM_EXCHANGE",
            index=1,
            number=1,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="JOB_REQUEST",
            index=2,
            number=2,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="JOB_ACCEPT",
            index=3,
            number=3,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="JOB_REJECT",
            index=4,
            number=4,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TRAINING_UPDATE",
            index=5,
            number=5,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TRAINING_COMPLETE",
            index=6,
            number=6,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="EVALUATION_RESULTS",
            index=7,
            number=7,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="SECURE_SHARE",
            index=8,
            number=8,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="MODEL_PARAMETERS",
            index=9,
            number=9,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TASK_ABORT",
            index=10,
            number=10,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="LARGE_OBJECT_STORAGE_REQUEST",
            index=11,
            number=11,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="KEY_EXCHANGE",
            index=12,
            number=12,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="POD_DATA",
            index=13,
            number=13,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="SAML_REQUEST",
            index=14,
            number=14,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="SAML_RESPONSE",
            index=15,
            number=15,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="LOG_MESSAGE",
            index=16,
            number=16,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TRAINING_METRICS",
            index=17,
            number=17,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TASK_COMPLETE",
            index=18,
            number=18,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="OIDC_CHALLENGE",
            index=19,
            number=19,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="OIDC_AFC_PKCE_RESPONSE",
            index=20,
            number=20,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="OIDC_DEVICE_CODE_RESPONSE",
            index=21,
            number=21,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="ONLINE_CHECK",
            index=22,
            number=22,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="ONLINE_RESPONSE",
            index=23,
            number=23,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="PSI_DATASET",
            index=24,
            number=24,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TASK_START",
            index=25,
            number=25,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=965,
    serialized_end=1502,
)
_sym_db.RegisterEnumDescriptor(_BITFOUNTMESSAGE_MESSAGETYPE)


_PODDATA = _descriptor.Descriptor(
    name="PodData",
    full_name="messages.PodData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="podName",
            full_name="messages.PodData.podName",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="processor",
            full_name="messages.PodData.processor",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="podOS",
            full_name="messages.PodData.podOS",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="cpuCount",
            full_name="messages.PodData.cpuCount",
            index=3,
            number=4,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="totalMemoryBytes",
            full_name="messages.PodData.totalMemoryBytes",
            index=4,
            number=5,
            type=4,
            cpp_type=4,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="gpuCount",
            full_name="messages.PodData.gpuCount",
            index=5,
            number=6,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="gpuName",
            full_name="messages.PodData.gpuName",
            index=6,
            number=7,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_gpuName",
            full_name="messages.PodData._gpuName",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=65,
    serialized_end=221,
)


_BITFOUNTTASK = _descriptor.Descriptor(
    name="BitfountTask",
    full_name="messages.BitfountTask",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="podIdentifier",
            full_name="messages.BitfountTask.podIdentifier",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="encryptedTask",
            full_name="messages.BitfountTask.encryptedTask",
            index=1,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="taskURL",
            full_name="messages.BitfountTask.taskURL",
            index=2,
            number=4,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_encryptedTask",
            full_name="messages.BitfountTask._encryptedTask",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_taskURL",
            full_name="messages.BitfountTask._taskURL",
            index=1,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=223,
    serialized_end=340,
)


_BITFOUNTTASKS = _descriptor.Descriptor(
    name="BitfountTasks",
    full_name="messages.BitfountTasks",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="tasks",
            full_name="messages.BitfountTasks.tasks",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=342,
    serialized_end=396,
)


_COMMUNICATIONDETAILS_PODMAILBOXIDSENTRY = _descriptor.Descriptor(
    name="PodMailboxIdsEntry",
    full_name="messages.CommunicationDetails.PodMailboxIdsEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="messages.CommunicationDetails.PodMailboxIdsEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="messages.CommunicationDetails.PodMailboxIdsEntry.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=532,
    serialized_end=584,
)

_COMMUNICATIONDETAILS = _descriptor.Descriptor(
    name="CommunicationDetails",
    full_name="messages.CommunicationDetails",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="mailboxId",
            full_name="messages.CommunicationDetails.mailboxId",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="podMailboxIds",
            full_name="messages.CommunicationDetails.podMailboxIds",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="taskId",
            full_name="messages.CommunicationDetails.taskId",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[
        _COMMUNICATIONDETAILS_PODMAILBOXIDSENTRY,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=399,
    serialized_end=584,
)


_BITFOUNTMESSAGE_PODMAILBOXIDSENTRY = _descriptor.Descriptor(
    name="PodMailboxIdsEntry",
    full_name="messages.BitfountMessage.PodMailboxIdsEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="messages.BitfountMessage.PodMailboxIdsEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="messages.BitfountMessage.PodMailboxIdsEntry.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=532,
    serialized_end=584,
)

_BITFOUNTMESSAGE = _descriptor.Descriptor(
    name="BitfountMessage",
    full_name="messages.BitfountMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="messageType",
            full_name="messages.BitfountMessage.messageType",
            index=0,
            number=1,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="body",
            full_name="messages.BitfountMessage.body",
            index=1,
            number=2,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="recipient",
            full_name="messages.BitfountMessage.recipient",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="recipientMailboxId",
            full_name="messages.BitfountMessage.recipientMailboxId",
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="sender",
            full_name="messages.BitfountMessage.sender",
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="senderMailboxId",
            full_name="messages.BitfountMessage.senderMailboxId",
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="timestamp",
            full_name="messages.BitfountMessage.timestamp",
            index=6,
            number=7,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="podMailboxIds",
            full_name="messages.BitfountMessage.podMailboxIds",
            index=7,
            number=8,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="receiptHandle",
            full_name="messages.BitfountMessage.receiptHandle",
            index=8,
            number=9,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="taskId",
            full_name="messages.BitfountMessage.taskId",
            index=9,
            number=10,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[
        _BITFOUNTMESSAGE_PODMAILBOXIDSENTRY,
    ],
    enum_types=[
        _BITFOUNTMESSAGE_MESSAGETYPE,
    ],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_sender",
            full_name="messages.BitfountMessage._sender",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_receiptHandle",
            full_name="messages.BitfountMessage._receiptHandle",
            index=1,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_taskId",
            full_name="messages.BitfountMessage._taskId",
            index=2,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=587,
    serialized_end=1542,
)


_ACKNOWLEDGEMENT = _descriptor.Descriptor(
    name="Acknowledgement",
    full_name="messages.Acknowledgement",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="mailboxId",
            full_name="messages.Acknowledgement.mailboxId",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="receiptHandle",
            full_name="messages.Acknowledgement.receiptHandle",
            index=1,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="deleteMailbox",
            full_name="messages.Acknowledgement.deleteMailbox",
            index=2,
            number=4,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_deleteMailbox",
            full_name="messages.Acknowledgement._deleteMailbox",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=1544,
    serialized_end=1649,
)


_SUCCESSRESPONSE = _descriptor.Descriptor(
    name="SuccessResponse",
    full_name="messages.SuccessResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1651,
    serialized_end=1668,
)


_LARGESTORAGEREQUEST = _descriptor.Descriptor(
    name="LargeStorageRequest",
    full_name="messages.LargeStorageRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="senderMailboxId",
            full_name="messages.LargeStorageRequest.senderMailboxId",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="podName",
            full_name="messages.LargeStorageRequest.podName",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="contentSize",
            full_name="messages.LargeStorageRequest.contentSize",
            index=2,
            number=3,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="taskId",
            full_name="messages.LargeStorageRequest.taskId",
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_podName",
            full_name="messages.LargeStorageRequest._podName",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_taskId",
            full_name="messages.LargeStorageRequest._taskId",
            index=1,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=1671,
    serialized_end=1804,
)


_LARGESTORAGETASKREQUEST = _descriptor.Descriptor(
    name="LargeStorageTaskRequest",
    full_name="messages.LargeStorageTaskRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="podName",
            full_name="messages.LargeStorageTaskRequest.podName",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="contentSize",
            full_name="messages.LargeStorageTaskRequest.contentSize",
            index=1,
            number=2,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="taskId",
            full_name="messages.LargeStorageTaskRequest.taskId",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_taskId",
            full_name="messages.LargeStorageTaskRequest._taskId",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=1806,
    serialized_end=1901,
)


_BLOBSTORAGEDATA_UPLOADFIELDSENTRY = _descriptor.Descriptor(
    name="UploadFieldsEntry",
    full_name="messages.BlobStorageData.UploadFieldsEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="messages.BlobStorageData.UploadFieldsEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="messages.BlobStorageData.UploadFieldsEntry.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=2030,
    serialized_end=2081,
)

_BLOBSTORAGEDATA = _descriptor.Descriptor(
    name="BlobStorageData",
    full_name="messages.BlobStorageData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="uploadUrl",
            full_name="messages.BlobStorageData.uploadUrl",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="downloadUrl",
            full_name="messages.BlobStorageData.downloadUrl",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="uploadFields",
            full_name="messages.BlobStorageData.uploadFields",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[
        _BLOBSTORAGEDATA_UPLOADFIELDSENTRY,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1904,
    serialized_end=2081,
)

_PODDATA.oneofs_by_name["_gpuName"].fields.append(_PODDATA.fields_by_name["gpuName"])
_PODDATA.fields_by_name["gpuName"].containing_oneof = _PODDATA.oneofs_by_name[
    "_gpuName"
]
_BITFOUNTTASK.oneofs_by_name["_encryptedTask"].fields.append(
    _BITFOUNTTASK.fields_by_name["encryptedTask"]
)
_BITFOUNTTASK.fields_by_name[
    "encryptedTask"
].containing_oneof = _BITFOUNTTASK.oneofs_by_name["_encryptedTask"]
_BITFOUNTTASK.oneofs_by_name["_taskURL"].fields.append(
    _BITFOUNTTASK.fields_by_name["taskURL"]
)
_BITFOUNTTASK.fields_by_name["taskURL"].containing_oneof = _BITFOUNTTASK.oneofs_by_name[
    "_taskURL"
]
_BITFOUNTTASKS.fields_by_name["tasks"].message_type = _BITFOUNTTASK
_COMMUNICATIONDETAILS_PODMAILBOXIDSENTRY.containing_type = _COMMUNICATIONDETAILS
_COMMUNICATIONDETAILS.fields_by_name[
    "podMailboxIds"
].message_type = _COMMUNICATIONDETAILS_PODMAILBOXIDSENTRY
_BITFOUNTMESSAGE_PODMAILBOXIDSENTRY.containing_type = _BITFOUNTMESSAGE
_BITFOUNTMESSAGE.fields_by_name["messageType"].enum_type = _BITFOUNTMESSAGE_MESSAGETYPE
_BITFOUNTMESSAGE.fields_by_name[
    "podMailboxIds"
].message_type = _BITFOUNTMESSAGE_PODMAILBOXIDSENTRY
_BITFOUNTMESSAGE_MESSAGETYPE.containing_type = _BITFOUNTMESSAGE
_BITFOUNTMESSAGE.oneofs_by_name["_sender"].fields.append(
    _BITFOUNTMESSAGE.fields_by_name["sender"]
)
_BITFOUNTMESSAGE.fields_by_name[
    "sender"
].containing_oneof = _BITFOUNTMESSAGE.oneofs_by_name["_sender"]
_BITFOUNTMESSAGE.oneofs_by_name["_receiptHandle"].fields.append(
    _BITFOUNTMESSAGE.fields_by_name["receiptHandle"]
)
_BITFOUNTMESSAGE.fields_by_name[
    "receiptHandle"
].containing_oneof = _BITFOUNTMESSAGE.oneofs_by_name["_receiptHandle"]
_BITFOUNTMESSAGE.oneofs_by_name["_taskId"].fields.append(
    _BITFOUNTMESSAGE.fields_by_name["taskId"]
)
_BITFOUNTMESSAGE.fields_by_name[
    "taskId"
].containing_oneof = _BITFOUNTMESSAGE.oneofs_by_name["_taskId"]
_ACKNOWLEDGEMENT.oneofs_by_name["_deleteMailbox"].fields.append(
    _ACKNOWLEDGEMENT.fields_by_name["deleteMailbox"]
)
_ACKNOWLEDGEMENT.fields_by_name[
    "deleteMailbox"
].containing_oneof = _ACKNOWLEDGEMENT.oneofs_by_name["_deleteMailbox"]
_LARGESTORAGEREQUEST.oneofs_by_name["_podName"].fields.append(
    _LARGESTORAGEREQUEST.fields_by_name["podName"]
)
_LARGESTORAGEREQUEST.fields_by_name[
    "podName"
].containing_oneof = _LARGESTORAGEREQUEST.oneofs_by_name["_podName"]
_LARGESTORAGEREQUEST.oneofs_by_name["_taskId"].fields.append(
    _LARGESTORAGEREQUEST.fields_by_name["taskId"]
)
_LARGESTORAGEREQUEST.fields_by_name[
    "taskId"
].containing_oneof = _LARGESTORAGEREQUEST.oneofs_by_name["_taskId"]
_LARGESTORAGETASKREQUEST.oneofs_by_name["_taskId"].fields.append(
    _LARGESTORAGETASKREQUEST.fields_by_name["taskId"]
)
_LARGESTORAGETASKREQUEST.fields_by_name[
    "taskId"
].containing_oneof = _LARGESTORAGETASKREQUEST.oneofs_by_name["_taskId"]
_BLOBSTORAGEDATA_UPLOADFIELDSENTRY.containing_type = _BLOBSTORAGEDATA
_BLOBSTORAGEDATA.fields_by_name[
    "uploadFields"
].message_type = _BLOBSTORAGEDATA_UPLOADFIELDSENTRY
DESCRIPTOR.message_types_by_name["PodData"] = _PODDATA
DESCRIPTOR.message_types_by_name["BitfountTask"] = _BITFOUNTTASK
DESCRIPTOR.message_types_by_name["BitfountTasks"] = _BITFOUNTTASKS
DESCRIPTOR.message_types_by_name["CommunicationDetails"] = _COMMUNICATIONDETAILS
DESCRIPTOR.message_types_by_name["BitfountMessage"] = _BITFOUNTMESSAGE
DESCRIPTOR.message_types_by_name["Acknowledgement"] = _ACKNOWLEDGEMENT
DESCRIPTOR.message_types_by_name["SuccessResponse"] = _SUCCESSRESPONSE
DESCRIPTOR.message_types_by_name["LargeStorageRequest"] = _LARGESTORAGEREQUEST
DESCRIPTOR.message_types_by_name["LargeStorageTaskRequest"] = _LARGESTORAGETASKREQUEST
DESCRIPTOR.message_types_by_name["BlobStorageData"] = _BLOBSTORAGEDATA
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PodData = _reflection.GeneratedProtocolMessageType(
    "PodData",
    (_message.Message,),
    {
        "DESCRIPTOR": _PODDATA,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.PodData)
    },
)
_sym_db.RegisterMessage(PodData)

BitfountTask = _reflection.GeneratedProtocolMessageType(
    "BitfountTask",
    (_message.Message,),
    {
        "DESCRIPTOR": _BITFOUNTTASK,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.BitfountTask)
    },
)
_sym_db.RegisterMessage(BitfountTask)

BitfountTasks = _reflection.GeneratedProtocolMessageType(
    "BitfountTasks",
    (_message.Message,),
    {
        "DESCRIPTOR": _BITFOUNTTASKS,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.BitfountTasks)
    },
)
_sym_db.RegisterMessage(BitfountTasks)

CommunicationDetails = _reflection.GeneratedProtocolMessageType(
    "CommunicationDetails",
    (_message.Message,),
    {
        "PodMailboxIdsEntry": _reflection.GeneratedProtocolMessageType(
            "PodMailboxIdsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _COMMUNICATIONDETAILS_PODMAILBOXIDSENTRY,
                "__module__": "bitfount.federated.transport.protos.messages_pb2"
                # @@protoc_insertion_point(class_scope:messages.CommunicationDetails.PodMailboxIdsEntry)
            },
        ),
        "DESCRIPTOR": _COMMUNICATIONDETAILS,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.CommunicationDetails)
    },
)
_sym_db.RegisterMessage(CommunicationDetails)
_sym_db.RegisterMessage(CommunicationDetails.PodMailboxIdsEntry)

BitfountMessage = _reflection.GeneratedProtocolMessageType(
    "BitfountMessage",
    (_message.Message,),
    {
        "PodMailboxIdsEntry": _reflection.GeneratedProtocolMessageType(
            "PodMailboxIdsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _BITFOUNTMESSAGE_PODMAILBOXIDSENTRY,
                "__module__": "bitfount.federated.transport.protos.messages_pb2"
                # @@protoc_insertion_point(class_scope:messages.BitfountMessage.PodMailboxIdsEntry)
            },
        ),
        "DESCRIPTOR": _BITFOUNTMESSAGE,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.BitfountMessage)
    },
)
_sym_db.RegisterMessage(BitfountMessage)
_sym_db.RegisterMessage(BitfountMessage.PodMailboxIdsEntry)

Acknowledgement = _reflection.GeneratedProtocolMessageType(
    "Acknowledgement",
    (_message.Message,),
    {
        "DESCRIPTOR": _ACKNOWLEDGEMENT,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.Acknowledgement)
    },
)
_sym_db.RegisterMessage(Acknowledgement)

SuccessResponse = _reflection.GeneratedProtocolMessageType(
    "SuccessResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _SUCCESSRESPONSE,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.SuccessResponse)
    },
)
_sym_db.RegisterMessage(SuccessResponse)

LargeStorageRequest = _reflection.GeneratedProtocolMessageType(
    "LargeStorageRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _LARGESTORAGEREQUEST,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.LargeStorageRequest)
    },
)
_sym_db.RegisterMessage(LargeStorageRequest)

LargeStorageTaskRequest = _reflection.GeneratedProtocolMessageType(
    "LargeStorageTaskRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _LARGESTORAGETASKREQUEST,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.LargeStorageTaskRequest)
    },
)
_sym_db.RegisterMessage(LargeStorageTaskRequest)

BlobStorageData = _reflection.GeneratedProtocolMessageType(
    "BlobStorageData",
    (_message.Message,),
    {
        "UploadFieldsEntry": _reflection.GeneratedProtocolMessageType(
            "UploadFieldsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _BLOBSTORAGEDATA_UPLOADFIELDSENTRY,
                "__module__": "bitfount.federated.transport.protos.messages_pb2"
                # @@protoc_insertion_point(class_scope:messages.BlobStorageData.UploadFieldsEntry)
            },
        ),
        "DESCRIPTOR": _BLOBSTORAGEDATA,
        "__module__": "bitfount.federated.transport.protos.messages_pb2"
        # @@protoc_insertion_point(class_scope:messages.BlobStorageData)
    },
)
_sym_db.RegisterMessage(BlobStorageData)
_sym_db.RegisterMessage(BlobStorageData.UploadFieldsEntry)


_COMMUNICATIONDETAILS_PODMAILBOXIDSENTRY._options = None
_BITFOUNTMESSAGE_PODMAILBOXIDSENTRY._options = None
_BLOBSTORAGEDATA_UPLOADFIELDSENTRY._options = None

_MESSAGESERVICE = _descriptor.ServiceDescriptor(
    name="MessageService",
    full_name="messages.MessageService",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=2084,
    serialized_end=2568,
    methods=[
        _descriptor.MethodDescriptor(
            name="PodConnect",
            full_name="messages.MessageService.PodConnect",
            index=0,
            containing_service=None,
            input_type=_PODDATA,
            output_type=_SUCCESSRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="SetupTaskMailboxes",
            full_name="messages.MessageService.SetupTaskMailboxes",
            index=1,
            containing_service=None,
            input_type=_BITFOUNTTASKS,
            output_type=_COMMUNICATIONDETAILS,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="AcknowledgeMessage",
            full_name="messages.MessageService.AcknowledgeMessage",
            index=2,
            containing_service=None,
            input_type=_ACKNOWLEDGEMENT,
            output_type=_SUCCESSRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="GetBitfountMessage",
            full_name="messages.MessageService.GetBitfountMessage",
            index=3,
            containing_service=None,
            input_type=_COMMUNICATIONDETAILS,
            output_type=_BITFOUNTMESSAGE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="SendBitfountMessage",
            full_name="messages.MessageService.SendBitfountMessage",
            index=4,
            containing_service=None,
            input_type=_BITFOUNTMESSAGE,
            output_type=_SUCCESSRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="GetLargeObjectStorage",
            full_name="messages.MessageService.GetLargeObjectStorage",
            index=5,
            containing_service=None,
            input_type=_LARGESTORAGEREQUEST,
            output_type=_BLOBSTORAGEDATA,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_MESSAGESERVICE)

DESCRIPTOR.services_by_name["MessageService"] = _MESSAGESERVICE

# @@protoc_insertion_point(module_scope)
